"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

preprocess_checks.py
~~~~~~~~~~~~~~~~~~~~

Checks for model consistency and possible errors during preprocessing.

"""

import os
import logging

import numpy as np
import pandas as pd

from inspect import signature

import calliope
from calliope._version import __version__
from calliope.core.attrdict import AttrDict
from calliope.core.preprocess.util import get_all_carriers
from calliope.core.util.tools import load_function

logger = logging.getLogger(__name__)

DEFAULTS = AttrDict.from_yaml(os.path.join(os.path.dirname(calliope.__file__), 'config', 'defaults.yaml'))
POSSIBLE_COSTS = [i for i in DEFAULTS.techs.default_tech.costs.default_cost.keys()]


def check_overrides(config_model, override):
    """
    Perform checks on the override dict and override file inputs to ensure they
    are not doing something silly.
    """
    model_warnings = []
    info = []
    for key in override.as_dict_flat().keys():
        if key in config_model.as_dict_flat().keys():
            info.append(
                'Override applied to {}: {} -> {}'
                .format(key, config_model.get_key(key), override.get_key(key))
            )
        else:
            info.append(
                '`{}`:{} applied from override as new configuration'
                .format(key, override.get_key(key))
            )

    # Check if overriding coordinates are in the same coordinate system. If not,
    # delete all incumbent coordinates, ready for the new coordinates to come in
    if (any(['coordinates' in k for k in config_model.as_dict_flat().keys()]) and
            any(['coordinates' in k for k in override.as_dict_flat().keys()])):

        # get keys that might be deleted and incumbent coordinate system
        config_keys = [k for k in config_model.as_dict_flat().keys() if 'coordinates.' in k]
        config_coordinates = set([k.split('coordinates.')[-1] for k in config_keys])

        # get overriding coordinate system
        override_coordinates = set(
            k.split('coordinates.')[-1] for k in override.as_dict_flat().keys()
            if 'coordinates.' in k
        )

        # compare overriding and incumbent, deleting incumbent if overriding is different
        if config_coordinates != override_coordinates:
            for key in config_keys:
                config_model.del_key(key)
            model_warnings.append(
                'Updated from coordinate system {} to {}, using overrides'
                .format(config_coordinates, override_coordinates)
            )

    if info:
        logger.info('\n'.join(info))

    return model_warnings


def check_initial(config_model):
    """
    Perform initial checks of model and run config dicts.

    Returns
    -------
    model_warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    errors = []
    model_warnings = []

    # Check for version mismatch
    model_version = config_model.model.get('calliope_version', False)
    if model_version:
        if not str(model_version) in __version__:
            model_warnings.append(
                'Model configuration specifies calliope_version={}, '
                'but you are running {}. Proceed with caution!'.format(
                    model_version, __version__)
            )

    # Check top-level keys
    for k in config_model.keys():
        if k not in [
                'model', 'run', 'locations', 'tech_groups', 'techs', 'links',
                'overrides', 'scenarios', 'config_path', 'group_constraints']:
            model_warnings.append(
                'Unrecognised top-level configuration item: {}'.format(k)
            )

    # Check that all required top-level keys are specified
    for k in ['model', 'run', 'locations', 'techs']:
        if k not in config_model.keys():
            errors.append(
                'Model is missing required top-level configuration item: {}'.format(k)
            )

    # Check run configuration
    # Exclude solver_options and objective_options.cost_class from checks,
    # as we don't know all possible options for all solvers
    for k in config_model['run'].keys_nested():
        if (k not in DEFAULTS['run'].keys_nested() and
                'solver_options' not in k and 'objective_options.cost_class' not in k):
            model_warnings.append(
                'Unrecognised setting in run configuration: {}'.format(k)
            )

    # Check model configuration, but top-level keys only
    for k in config_model['model'].keys():
        if k not in DEFAULTS['model'].keys():
            model_warnings.append(
                'Unrecognised setting in model configuration: {}'.format(k)
            )

    # Only ['in', 'out', 'in_2', 'out_2', 'in_3', 'out_3']
    # are allowed as carrier tiers
    for key in config_model.as_dict_flat().keys():
        if ('essentials.carrier_' in key
                and key.split('.carrier_')[-1].split('.')[0] not in
                ['in', 'out', 'in_2', 'out_2', 'in_3', 'out_3']):
            errors.append(
                "Invalid carrier tier found at {}. Only "
                "'carrier_' + ['in', 'out', 'in_2', 'out_2', 'in_3', 'out_3'] "
                "is valid.".format(key)
            )

    # Warn if any unknown group constraints are defined
    permitted_group_constraints = DEFAULTS.group_constraints.default_group.keys()

    for group in config_model.get('group_constraints', {}).keys():
        for key in config_model.group_constraints[group].keys():
            if key not in permitted_group_constraints:
                model_warnings.append(
                    'Unrecognised group constraint `{}` in group `{}` '
                    'will be ignored - possibly a misspelling?'.format(key, group)
                )

    # No techs may have the same identifier as a tech_group
    name_overlap = (
        set(config_model.tech_groups.keys()) &
        set(config_model.techs.keys())
    )
    if name_overlap:
        errors.append(
            'tech_groups and techs with '
            'the same name exist: {}'.format(name_overlap)
        )

    # Checks for techs and tech_groups:
    # * All user-defined tech and tech_groups must specify a parent
    # * techs cannot be parents, only tech groups can
    # * No carrier may be called 'resource'
    default_tech_groups = list(DEFAULTS.tech_groups.keys())
    for tg_name, tg_config in config_model.tech_groups.items():
        if tg_name in default_tech_groups:
            continue
        if not tg_config.get_key('essentials.parent'):
            errors.append(
                'tech_group {} does not define '
                '`essentials.parent`'.format(tg_name)
            )
        elif tg_config.get_key('essentials.parent') in config_model.techs.keys():
            errors.append(
                'tech_group `{}` has a tech as a parent, only another tech_group '
                'is allowed'.format(tg_name)
            )
        if 'resource' in get_all_carriers(tg_config.essentials):
            errors.append(
                'No carrier called `resource` may '
                'be defined (tech_group: {})'.format(tg_name)
            )

    for t_name, t_config in config_model.techs.items():
        for key in t_config.keys():
            if key not in DEFAULTS.techs.default_tech.keys():
                model_warnings.append("Unknown key `{}` defined for tech {}.".format(key, t_name))
        if not t_config.get_key('essentials.parent'):
            errors.append(
                'tech {} does not define '
                '`essentials.parent`'.format(t_name)
            )
        elif t_config.get_key('essentials.parent') in config_model.techs.keys():
            errors.append(
                'tech `{}` has another tech as a parent, only a tech_group '
                'is allowed'.format(tg_name)
            )
        if 'resource' in get_all_carriers(t_config.essentials):
            errors.append(
                'No carrier called `resource` may '
                'be defined (tech: {})'.format(t_name)
            )

    # Check whether any unrecognised mid-level keys are defined in techs, locations, or links
    for k, v in config_model.get('locations', {}).items():
        unrecognised_keys = [
            i for i in v.keys()
            if i not in DEFAULTS.locations.default_location.keys()
        ]
        if len(unrecognised_keys) > 0:
            errors.append(
                'Location `{}` contains unrecognised keys {}. '
                'These could be mispellings or a technology not defined '
                'under the `techs` key.'.format(k, unrecognised_keys)
            )
        for loc_tech_key, loc_tech_val in v.get('techs', {}).items():
            if loc_tech_val is None:
                continue
            unrecognised_keys = [
                i for i in loc_tech_val.keys()
                if i not in DEFAULTS.techs.default_tech.keys()
            ]
            if len(unrecognised_keys) > 0:
                errors.append(
                    'Technology `{}` in location `{}` contains unrecognised keys {}; '
                    'these are most likely mispellings'.format(loc_tech_key, k, unrecognised_keys)
                )

    default_link = DEFAULTS.links['default_location_from,default_location_to']
    for k, v in config_model.get('links', {}).items():
        unrecognised_keys = [
            i for i in v.keys()
            if i not in default_link.keys()
        ]
        if len(unrecognised_keys) > 0:
            errors.append(
                'Link `{}` contains unrecognised keys {}. '
                'These could be mispellings or a technology not defined '
                'under the `techs` key.'.format(k, unrecognised_keys)
            )
        for link_tech_key, link_tech_val in v.get('techs', {}).items():
            if link_tech_val is None:
                continue
            unrecognised_keys = [
                i for i in link_tech_val.keys()
                if i not in default_link.techs.default_tech.keys()
                and i not in DEFAULTS.techs.default_tech.keys()
            ]
            if len(unrecognised_keys) > 0:
                errors.append(
                    'Technology `{}` in link `{}` contains unrecognised keys {}; '
                    'these are most likely mispellings'.format(link_tech_key, k, unrecognised_keys)
                )

    # Error if a technology is defined twice, in opposite directions
    link_techs = [
        tuple(sorted(j.strip() for j in k.split(','))) + (i, )
        for k, v in config_model.get('links', {}).items()
        for i in v.get('techs', {}).keys()
    ]
    if len(link_techs) != len(set(link_techs)):
        duplicated_techs = np.array(link_techs)[pd.Series(link_techs).duplicated().values]
        duplicated_techs = set([i[-1] for i in duplicated_techs])
        tech_end = 'y' if len(duplicated_techs) == 1 else 'ies'
        errors.append(
            'Technolog{} {} defined twice on a link defined in both directions '
            '(e.g. `A,B` and `B,A`). A technology can only be defined on one link '
            'even if it allows unidirectional flow in each direction '
            '(i.e. `one_way: true`).'.format(tech_end, ', '.join(duplicated_techs))
        )

    # Error if a constraint is loaded from file that must not be
    allowed_from_file = DEFAULTS.model.file_allowed
    for k, v in config_model.as_dict_flat().items():
        if 'file=' in str(v):

            possible_identifiers = k.split('.')
            is_time_varying = any('_time_varying' in i for i in possible_identifiers)
            if is_time_varying:
                model_warnings.append('Using custom constraint `{}` with time-varying data.'.format(k))
            elif not set(possible_identifiers).intersection(allowed_from_file) and not is_time_varying:
                errors.append(
                    'Cannot load data from file for configuration `{}`.'.format(k)
                )

    # We no longer allow cost_class in objective_obtions to be a string
    _cost_class = config_model.run.objective_options.get('cost_class', {})

    if not isinstance(_cost_class, dict):
        errors.append(
            '`run.objective_options.cost_class` must be a dictionary.'
            'If you want to minimise or maximise with a single cost class, '
            'use e.g. "{monetary: 1}", which gives the monetary cost class a weight '
            'of 1 in the objective, and ignores any other cost classes.'
        )
    elif len(_cost_class.keys()) == 0:
        errors.append(
            'No cost classes defined for use in the objective. '
            'Expecting a dict of "{cost_class: weight}" for all cost classes '
            'to be considered in the objective function.'
        )

    else:
        # This next check is only run if we have confirmed that cost_class is
        # a dict, as it errors otherwise

        # For cost minimisation objective, check for cost_class: None and set to one
        for k, v in _cost_class.items():
            if v is None:
                config_model.run.objective_options.cost_class[k] = 1
                model_warnings.append(
                    'cost class {} has weight = None, setting weight to 1'.format(k)
                )

    if (isinstance(_cost_class, dict) and _cost_class.get('monetary', 0) == 1
            and len(_cost_class.keys()) > 1):
        # Warn that {monetary: 1} is still in the objective, since it is not
        # automatically overidden on setting another objective.
        model_warnings.append(
            'Monetary cost class with a weight of 1 is still included '
            'in the objective. If you want to remove the monetary cost class, '
            'add `{"monetary": 0}` to the dictionary nested under '
            ' `run.objective_options.cost_class`.'
        )

    # Don't allow time clustering with cyclic storage if not also using
    # storage_inter_cluster
    storage_inter_cluster = 'model.time.function_options.storage_inter_cluster'
    if (config_model.get_key('model.time.function', None) == 'apply_clustering'
            and config_model.get_key('run.cyclic_storage', True)
            and not config_model.get_key(storage_inter_cluster, True)):
        errors.append(
            'When time clustering, cannot have cyclic storage constraints if '
            '`storage_inter_cluster` decision variable is not activated.'
        )

    return model_warnings, errors


def _check_tech_final(model_run, tech_id, tech_config, loc_id, model_warnings, errors, comments):
    """
    Checks individual tech/tech groups at specific locations.
    NOTE: Updates `model_warnings` and `errors` lists in-place.
    """
    if tech_id not in model_run.techs:
        model_warnings.append(
            'Tech {} was removed by setting ``exists: False`` - not checking '
            'the consistency of its constraints at location {}.'.format(tech_id, loc_id)
        )
        return model_warnings, errors

    required = model_run.techs[tech_id].required_constraints
    allowed = model_run.techs[tech_id].allowed_constraints
    allowed_costs = model_run.techs[tech_id].allowed_costs

    # Error if required constraints are not defined
    for r in required:
        # If it's a string, it must be defined
        single_ok = isinstance(r, str) and r in tech_config.constraints
        # If it's a list of strings, one of them must be defined
        multiple_ok = (
            isinstance(r, list) and
            any([i in tech_config.constraints for i in r])
        )
        if not single_ok and not multiple_ok:
            errors.append(
                '`{}` at `{}` fails to define '
                'all required constraints: {}'.format(tech_id, loc_id, required)
            )

    # Warn if defining a carrier ratio for a conversion_plus tech,
    # but applying it to a carrier that isn't one of the carriers specified by that tech
    # e.g. carrier_ratios.carrier_in_2.cooling when cooling isn't a carrier`
    defined_carriers = get_all_carriers(model_run.techs[tech_id].essentials)
    carriers_in_ratios = [
        i.split('.')[-1] for i in
        tech_config.constraints.get_key('carrier_ratios', AttrDict()).as_dict_flat().keys()
    ]
    for carrier in carriers_in_ratios:
        if carrier not in defined_carriers:
            model_warnings.append(
                'Tech `{t}` gives a carrier ratio for `{c}`, but does not actually '
                'configure `{c}` as a carrier.'.format(t=tech_id, c=carrier)
            )

    # If the technology involves storage, warn when energy_cap and storage_cap aren't connected
    energy_cap_per_storage_cap_params = [
        'charge_rate', 'energy_cap_per_storage_cap_min',
        'energy_cap_per_storage_cap_max', 'energy_cap_per_storage_cap_equals'
    ]
    if (loc_id + '::' + tech_id in model_run.sets.loc_techs_store
            and not any(i in tech_config.constraints.keys() for i in energy_cap_per_storage_cap_params)):
        logger.info(
            '`{}` at `{}` has no constraint to explicitly connect `energy_cap` to '
            '`storage_cap`, consider defining a `energy_cap_per_storage_cap_min/max/equals` '
            'constraint'.format(tech_id, loc_id)
        )

    # If a technology is defined by units (i.e. integer decision variable), it must define energy_cap_per_unit
    if (any(['units_' in k for k in tech_config.constraints.keys()])
            and 'energy_cap_per_unit' not in tech_config.constraints.keys()):
        errors.append(
            '`{}` at `{}` fails to define energy_cap_per_unit when specifying '
            'technology in units_max/min/equals'.format(tech_id, loc_id, required)
        )

    # If a technology is defined by units & is a storage tech, it must define storage_cap_per_unit
    if (any(['units_' in k for k in tech_config.constraints.keys()])
            and model_run.techs[tech_id].essentials.parent in ['storage', 'supply_plus']
            and any(['storage' in k for k in tech_config.constraints.keys()])
            and 'storage_cap_per_unit' not in tech_config.constraints.keys()):
        errors.append(
            '`{}` at `{}` fails to define storage_cap_per_unit when specifying '
            'technology in units_max/min/equals'.format(tech_id, loc_id, required)
        )

    # If a technology defines force_resource but is not in loc_techs_finite_resource
    if ('force_resource' in tech_config.constraints.keys() and
            loc_id + '::' + tech_id not in model_run.sets.loc_techs_finite_resource):

        model_warnings.append(
            '`{}` at `{}` defines force_resource but not a finite resource, so '
            'force_resource will not be applied'.format(tech_id, loc_id)
        )

    # Gather remaining unallowed constraints
    remaining = set(tech_config.constraints) - set(allowed)

    # Error if something is defined that's not allowed, but is in defaults
    # Warn if something is defined that's not allowed, but is not in defaults
    # (it could be a misspelling)
    for k in remaining:
        if k in DEFAULTS.techs.default_tech.constraints.keys():
            errors.append(
                '`{}` at `{}` defines non-allowed '
                'constraint `{}`'.format(tech_id, loc_id, k)
            )
        else:
            model_warnings.append(
                '`{}` at `{}` defines unrecognised '
                'constraint `{}` - possibly a misspelling?'.format(tech_id, loc_id, k)
            )

    # Error if an `export` statement does not match the given carrier_outs
    if 'export_carrier' in tech_config.constraints:
        essentials = model_run.techs[tech_id].essentials
        export = tech_config.constraints.export_carrier
        if (export and export not in [essentials.get_key(k, '')
                for k in ['carrier_out', 'carrier_out_2', 'carrier_out_3']]):
            errors.append(
                '`{}` at `{}` is attempting to export a carrier '
                'not given as an output carrier: `{}`'.format(tech_id, loc_id, export)
            )

    # Error if non-allowed costs are defined
    for cost_class in tech_config.get_key('costs', {}):
        for k in tech_config.costs[cost_class]:
            if k not in allowed_costs:
                errors.append(
                    '`{}` at `{}` defines non-allowed '
                    '{} cost: `{}`'.format(tech_id, loc_id, cost_class, k)
                )

    # Error if non-allowed `resource_unit` is defined
    if (tech_config.constraints.get('resource_unit', 'energy') not in
            ['energy', 'energy_per_cap', 'energy_per_area']):
        errors.append(
            '`{}` is an unknown resource unit for `{}` at `{}`. '
            'Only `energy`, `energy_per_cap`, or `energy_per_area` is allowed.'
            .format(tech_config.constraints.resource_unit, tech_id, loc_id)
        )

    return model_warnings, errors


def check_final(model_run):
    """
    Perform final checks of the completely built model_run.

    Returns
    -------
    comments : AttrDict
        debug output
    model_warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    model_warnings, errors = [], []
    comments = AttrDict()

    # Go through all loc-tech combinations and check validity
    for loc_id, loc_config in model_run.locations.items():
        if 'techs' in loc_config:
            for tech_id, tech_config in loc_config.techs.items():
                _check_tech_final(
                    model_run, tech_id, tech_config, loc_id,
                    model_warnings, errors, comments
                )

        if 'links' in loc_config:
            for link_id, link_config in loc_config.links.items():
                for tech_id, tech_config in link_config.techs.items():
                    _check_tech_final(
                        model_run, tech_id, tech_config,
                        'link {}:{}'.format(loc_id, link_id),
                        model_warnings, errors, comments
                    )

    # Either all locations or no locations must have coordinates
    all_locs = list(model_run.locations.keys())
    locs_with_coords = [
        k for k in model_run.locations.keys()
        if 'coordinates' in model_run.locations[k]
    ]
    if len(locs_with_coords) != 0 and len(all_locs) != len(locs_with_coords):
        errors.append(
            'Either all or no locations must have `coordinates` defined. '
            'Locations defined: {} - Locations with coordinates: {}'.format(
                all_locs, locs_with_coords)
        )

    # If locations have coordinates, they must all be either lat/lon or x/y
    elif len(locs_with_coords) != 0:
        first_loc = list(model_run.locations.keys())[0]
        try:
            coord_keys = sorted(list(model_run.locations[first_loc].coordinates.keys()))
            if coord_keys != ['lat', 'lon'] and coord_keys != ['x', 'y']:
                errors.append(
                    'Unidentified coordinate system. All locations must either'
                    'use the format {lat: N, lon: M} or {x: N, y: M}.'
                )
        except AttributeError:
            errors.append(
                'Coordinates must be given in the format {lat: N, lon: M} or '
                '{x: N, y: M}, not ' + str(model_run.locations[first_loc].coordinates)
            )

        for loc_id, loc_config in model_run.locations.items():
            try:
                if sorted(list(loc_config.coordinates.keys())) != coord_keys:
                    errors.append('All locations must use the same coordinate format.')
                    break
            except AttributeError:
                errors.append(
                    'Coordinates must be given in the format {lat: N, lon: M} or '
                    '{x: N, y: M}, not ' + str(model_run.locations[first_loc].coordinates)
                )
                break

    # Ensure that timeseries have no non-unique index values
    for k, df in model_run['timeseries_data'].items():
        if df.index.duplicated().any():
            errors.append('Time series `{}` contains non-unique timestamp values.'.format(k))

    # Warn if loc/tech is defined in group constraint that doesn't exist in the model
    for i in ['locs', 'techs']:
        config_i = 'locations' if i == 'locs' else i
        _missing = set()
        for group, group_vals in model_run.get('group_constraints', {}).items():
            if i in group_vals.keys() and set(group_vals[i]).difference(model_run[config_i].keys()):
                _missing.update(set(group_vals[i]).difference(model_run[config_i].keys()))
                model_run['group_constraints'][group][i] = (
                    list(set(group_vals[i]).intersection(model_run[config_i].keys()))
                )
                if model_run['group_constraints'][group][i] == []:
                    model_warnings.append(
                        'Constraint group `{}` will be completely ignored since '
                        'none of the defined {} are valid for this model.'
                        .format(group, i)
                    )
                    model_run['group_constraints'][group]['exists'] = False
        if _missing:
            model_warnings.append(
                'Possible misspelling in group constraints: {0} {1} given in '
                'group constraints, but not defined as {0} in the model. They '
                'will be ignored in the optimisation run'.format(i, _missing)
            )

    # Warn if a group constraint will ignore tech(s)
    group_constraints = {
        name: data for name, data in model_run['group_constraints'].items()
        if data.get("exists", True)
    }
    for group_constraint_name, group_constraint in group_constraints.items():
        requested_techs = [i for i in group_constraint.get('techs', [])]
        requested_locs = group_constraint.get('locs', model_run.sets.locs)

        allowed_tech_groups = [
            [k for k, v in DEFAULTS.tech_groups.items() if i in v['allowed_group_constraints']]
            for i in group_constraint.keys() if i not in ['techs', 'locs', 'exists']
        ]
        unallowed_tech_groups = (
            set([item for sublist in allowed_tech_groups for item in sublist])
            .symmetric_difference(DEFAULTS.tech_groups.keys())
        )

        stripped_allowed_tech_groups = set(allowed_tech_groups[0]).intersection(*allowed_tech_groups)
        group_loc_techs = model_run.constraint_sets['group_constraint_loc_techs_{}'.format(group_constraint_name)]
        dropped_tech_groups = (
            set([item for sublist in allowed_tech_groups for item in sublist])
            - stripped_allowed_tech_groups
        )

        if not group_loc_techs:
            model_warnings.append(
                'Group constraint `{}` cannot be applied as there are no valid loc_tech '
                'combinations available.'.format(group_constraint_name)
            )
            continue

        if not requested_techs:
            model_warnings.append(
                'All technologies were requested for inclusion in group constraint '
                '`{}`, but those from tech group(s) `{}` have been ignored as one '
                'or more of the constraints cannot be applied to technologies '
                'in these groups'.format(
                    group_constraint_name,
                    sorted(dropped_tech_groups | unallowed_tech_groups)
                )
            )
        else:
            trans_techs = set(requested_techs).intersection(model_run.sets['techs_transmission_names'])
            for i in trans_techs:
                requested_techs += [i + ':' + j for j in requested_locs]
                requested_techs.remove(i)
            group_techs = set(i.split('::')[1] for i in group_loc_techs)
            dropped_techs = set(requested_techs).difference(group_techs)
            if dropped_techs:
                _mismatch = [group_constraint_name, dropped_techs, unallowed_tech_groups, dropped_tech_groups]
                model_warnings.append(
                    'The following requested techs have been removed in group constraint '
                    '`{}`: {}. This has been caused by no constraints allowing '
                    'techs from tech group(s) {} and some constraints not '
                    'allowing techs from tech group(s) {}.'.format(*_mismatch)
                )

    # Warn if objective cost class is not defined elsewhere in the model
    objective_cost_class = set(model_run.run.objective_options.cost_class.keys())
    cost_classes = model_run.sets.costs
    cost_classes_mismatch = objective_cost_class.difference(cost_classes)
    if cost_classes_mismatch:
        model_warnings.append(
            'Cost classes `{}` are defined in the objective options but not '
            'defined elsewhere in the model. They will be ignored in the '
            'objective function.'.format(cost_classes_mismatch)
        )

    # FIXME:
    # make sure `comments` is at the the base level:
    # i.e. comments.model_run.xxxxx....

    return comments, model_warnings, errors


def check_model_data(model_data):
    """
    Perform final checks of the completely built xarray Dataset `model_data`.

    Returns
    -------
    comments : AttrDict
        debug output
    model_warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    model_warnings, errors = [], []
    comments = AttrDict()

    # Ensure that no loc-tech specifies infinite resource and force_resource=True
    if "force_resource" in model_data.data_vars:
        relevant_loc_techs = [
            i.loc_techs_finite_resource.item()
            for i in model_data.force_resource if i.item() is True
        ]
        forced_resource = model_data.resource.loc[
            dict(loc_techs_finite_resource=relevant_loc_techs)
        ]
        conflict = forced_resource.where(forced_resource == np.inf).to_pandas().dropna()
        if not conflict.empty:
            errors.append(
                'loc_tech(s) {} cannot have `force_resource` set as infinite '
                'resource values are given'.format(', '.join(conflict.index))
            )

    # Ensure that if a tech has negative costs, there is a max cap defined
    # FIXME: doesn't consider capacity being set by a linked constraint e.g.
    # `resource_cap_per_energy_cap`.
    relevant_caps = [
        i for i in ['energy_cap', 'storage_cap', 'resource_cap', 'resource_area']
        if 'cost_' + i in model_data.data_vars.keys()
    ]
    for cap in relevant_caps:
        relevant_loc_techs = (model_data['cost_' + cap]
                              .where(model_data['cost_' + cap] < 0, drop=True)
                              .to_pandas())
        cap_max = cap + '_max'
        cap_equals = cap + '_equals'
        for loc_tech in relevant_loc_techs.columns:
            try:
                cap_val = model_data[cap_max][loc_tech].item()
            except KeyError:
                try:
                    cap_val = model_data[cap_equals][loc_tech].item()
                except KeyError:
                    cap_val = np.nan
            if np.isinf(cap_val) or np.isnan(cap_val):
                errors.append(
                    'loc_tech {} cannot have a negative cost_{} as the '
                    'corresponding capacity constraint is not set'
                    .format(loc_tech, cap)
                )

    for loc_tech in set(model_data.loc_techs_demand.values).intersection(model_data.loc_techs_finite_resource.values):
        if any(model_data.resource.loc[loc_tech].values > 0):
            errors.append(
                'Positive resource given for demand loc_tech {}. All demands '
                'must have negative resource'.format(loc_tech)
            )

    # Delete all empty dimensions & the variables associated with them
    for dim_name, dim_length in model_data.dims.items():
        if dim_length == 0:
            if dim_name in model_data.coords.keys():
                del model_data[dim_name]
            associated_vars = [
                var for var, data in model_data.data_vars.items() if dim_name in data.dims
            ]
            model_data = model_data.drop(associated_vars)
            model_warnings.append(
                'dimension {} and associated variables {} were empty, so have '
                'been deleted'.format(dim_name, ', '.join(associated_vars))
            )

    # Check if we're allowed to use operate mode
    if 'allow_operate_mode' not in model_data.attrs.keys():
        daily_timesteps = [
            model_data.timestep_resolution.loc[i].values
            for i in np.unique(model_data.timesteps.to_index().strftime('%Y-%m-%d'))
        ]
        if not np.all(daily_timesteps == daily_timesteps[0]):
            model_data.attrs['allow_operate_mode'] = 0
            model_warnings.append(
                'Operational mode requires the same timestep resolution profile '
                'to be emulated on each date'
            )
        else:
            model_data.attrs['allow_operate_mode'] = 1

    # Check for any milp constraints, and warn that the problem contains binary /
    # integer decision variables
    if any('_milp_constraint' in i for i in model_data.dims):
        model_warnings.append(
            'Integer and / or binary decision variables are included in this model. '
            'This may adversely affect solution time, particularly if you are '
            'using a non-commercial solver. To improve solution time, consider '
            'changing MILP related solver options (e.g. `mipgap`) or removing '
            'MILP constraints.'
        )

    # Check for storage_initial being a fractional value

    if hasattr(model_data, 'loc_techs_store'):
        for loc_tech in model_data.loc_techs_store.values:
            if hasattr(model_data, 'storage_initial'):
                if model_data.storage_initial.loc[{'loc_techs_store': loc_tech}].values > 1:
                    errors.append(
                        'storage_initial values larger than 1 are not allowed.'
                    )
                if model_data.storage_initial.loc[{'loc_techs_store': loc_tech}].values == 0:
                    model_data.storage_initial.loc[{'loc_techs_store': loc_tech}] = 0.0

    # Check for storage_initial being greater than or equal to the storage_discharge_depth

    if hasattr(model_data, 'loc_techs_store'):
        for loc_tech in model_data.loc_techs_store.values:
            if hasattr(model_data, 'storage_initial') and hasattr(model_data, 'storage_discharge_depth'):
                if (model_data.storage_initial.loc[{'loc_techs_store': loc_tech}].values
                        < model_data.storage_discharge_depth.loc[{'loc_techs_store': loc_tech}].values):
                    errors.append(
                        'storage_initial is smaller than storage_discharge_depth.'
                        ' Please change the model configuration to ensure that'
                        ' storage initial is greater than or equal to storage_discharge_depth'
                    )

    # Check for storage_inter_cluster not being used together with storage_discharge_depth
    if hasattr(model_data, 'clusters') and hasattr(model_data, 'storage_discharge_depth'):
        errors.append(
            'storage_discharge_depth is currently not allowed when time clustering is active.'
        )

    return model_data, comments, model_warnings, errors



import logging

import xarray as xr
import pyomo.core as po  # pylint: disable=import-error

import calliope
from calliope.backend.pyomo.util import get_var
from calliope.backend import run as backend_run
from calliope.backend.pyomo import model as run_pyomo

from calliope.core.util.dataset import reorganise_xarray_dimensions
from calliope.core.util.logging import log_time
from calliope import exceptions
from calliope.core.attrdict import AttrDict
from calliope.analysis.postprocess import postprocess_model_results

logger = logging.getLogger(__name__)


def access_pyomo_model_inputs(backend_model):
    """
    If the user wishes to inspect the parameter values used as inputs in the backend
    model, they can access a new Dataset of all the backend model inputs, including
    defaults applied where the user did not specify anything for a loc::tech
    """

    all_params = {
        i.name: get_var(backend_model, i.name, sparse=True)
        for i in backend_model.component_objects()
        if isinstance(i, po.base.param.IndexedParam)
    }

    return reorganise_xarray_dimensions(xr.Dataset(all_params))


def update_pyomo_param(backend_model, param, update_dict):
    """
    A Pyomo Param value can be updated without the user directly accessing the
    backend model.

    Parameters
    ----------
    param : str
        Name of the parameter to update
    update_dict : dict
        keys are parameter indeces (either strings or tuples of strings,
        depending on whether there is one or more than one dimension). Values
        are the new values being assigned to the parameter at the given indeces.

    Returns
    -------
    Value(s) will be updated in-place, requiring the user to run the model again to
    see the effect on results.

    """
    if not hasattr(backend_model, param):
        raise exceptions.ModelError(
            'Parameter `{}` not in the Pyomo Backend. Check that the string '
            'matches the corresponding constraint/cost in the model.inputs '
            'xarray Dataset'.format(param)
        )
    elif not isinstance(getattr(backend_model, param), po.base.param.IndexedParam):
        raise exceptions.ModelError(
            '`{}` not a Parameter in the Pyomo Backend. Sets and decision variables '
            'cannot be updated by the user'.format(param)
        )
    elif not isinstance(update_dict, dict):
        raise TypeError('`update_dict` must be a dictionary')

    else:
        print(
            'Warning: we currently do not check that the updated value is the '
            'correct data type for this Pyomo Parameter, this is your '
            'responsibility to check!'
        )
        getattr(backend_model, param).store_values(update_dict)


def activate_pyomo_constraint(backend_model, constraint, active=True):
    """
    Takes a constraint or objective name, finds it in the backend model and sets
    its status to either active or deactive.

    Parameters
    ----------
    constraint : str
        Name of the constraint/objective to activate/deactivate
        Built-in constraints include '_constraint'
    active: bool, default=True
        status to set the constraint/objective
    """
    if not hasattr(backend_model, constraint):
        raise exceptions.ModelError(
            'constraint/objective `{}` not in the Pyomo Backend.'.format(constraint)
        )
    elif not isinstance(getattr(backend_model, constraint), po.base.Constraint):
        raise exceptions.ModelError(
            '`{}` not a constraint in the Pyomo Backend.'.format(constraint)
        )
    elif active is True:
        getattr(backend_model, constraint).activate()
    elif active is False:
        getattr(backend_model, constraint).deactivate()
    else:
        raise ValueError('Argument `active` must be True or False')


def rerun_pyomo_model(model_data, backend_model):
    """
    Rerun the Pyomo backend, perhaps after updating a parameter value,
    (de)activating a constraint/objective or updating run options in the model
    model_data object (e.g. `run.solver`).

    Returns
    -------
    new_model : calliope.Model
        New calliope model, including both inputs and results, but no backend interface.
    """
    backend_model.__calliope_run_config = AttrDict.from_yaml_string(model_data.attrs['run_config'])

    if backend_model.__calliope_run_config['mode'] != 'plan':
        raise exceptions.ModelError(
            'Cannot rerun the backend in {} run mode. Only `plan` mode is '
            'possible.'.format(backend_model.__calliope_run_config['mode'])
        )

    timings = {}
    log_time(logger, timings, 'model_creation')

    results, backend_model = backend_run.run_plan(
        model_data, timings, run_pyomo,
        build_only=False, backend_rerun=backend_model
    )

    inputs = access_pyomo_model_inputs(backend_model)

    # Add additional post-processed result variables to results
    if results.attrs.get('termination_condition', None) in ['optimal', 'feasible']:
        results = postprocess_model_results(
            results, model_data.reindex(results.coords), timings
        )

    for key, var in results.data_vars.items():
        var.attrs['is_result'] = 1

    for key, var in inputs.data_vars.items():
        var.attrs['is_result'] = 0

    new_model_data = xr.merge((results, inputs))
    new_model_data.attrs.update(model_data.attrs)

    # Only add coordinates from the original model_data that don't already exist
    new_coords = [
        i for i in model_data.coords.keys() if i not in new_model_data.coords.keys()
    ]
    new_model_data = new_model_data.update(model_data[new_coords])

    # Reorganise the coordinates so that model data and new model data share
    # the same order of items in each dimension
    new_model_data = new_model_data.reindex(model_data.coords)

    exceptions.warn(
        'The results of rerunning the backend model are only available within '
        'the Calliope model returned by this function call.'
    )

    new_calliope_model = calliope.Model(config=None, model_data=new_model_data)
    new_calliope_model._timings = timings

    return new_calliope_model


class BackendInterfaceMethods:

    def __init__(self, model):
        self._backend = model._backend_model
        self._model_data = model._model_data

    def access_model_inputs(self):
        return access_pyomo_model_inputs(self._backend)

    access_model_inputs.__doc__ = access_pyomo_model_inputs.__doc__

    def update_param(self, *args, **kwargs):
        return update_pyomo_param(self._backend, *args, **kwargs)

    update_param.__doc__ = update_pyomo_param.__doc__

    def activate_constraint(self, *args, **kwargs):
        return activate_pyomo_constraint(self._backend, *args, **kwargs)

    activate_constraint.__doc__ = activate_pyomo_constraint.__doc__

    def rerun(self, *args, **kwargs):
        return rerun_pyomo_model(self._model_data, self._backend, *args, **kwargs)

    rerun.__doc__ = rerun_pyomo_model.__doc__



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

io.py
~~~~~

Functions to read and save model results.

"""

import os

import xarray as xr

from calliope._version import __version__
from calliope import exceptions
from calliope.core.util.dataset import split_loc_techs


def read_netcdf(path):
    """Read model_data from NetCDF file"""
    with xr.open_dataset(path) as model_data:
        model_data.load()

    calliope_version = model_data.attrs.get('calliope_version', False)
    if calliope_version:
        if not str(calliope_version) in __version__:
            exceptions.warn(
                'This model data was created with Calliope version {}, '
                'but you are running {}. Proceed with caution!'.format(calliope_version, __version__)
            )

    # FIXME some checks for consistency
    # use check_dataset from the checks module
    # also check the old checking from 0.5.x

    return model_data


def save_netcdf(model_data, path, model=None):
    encoding = {k: {'zlib': True, 'complevel': 4} for k in model_data.data_vars}

    original_model_data_attrs = model_data.attrs
    model_data_attrs = model_data.attrs.copy()

    if model is not None and hasattr(model, '_model_run'):
        # Attach _model_run and _debug_data to _model_data
        model_run_to_save = model._model_run.copy()
        if 'timeseries_data' in model_run_to_save:
            del model_run_to_save['timeseries_data']  # Can't be serialised!
        model_data_attrs['_model_run'] = model_run_to_save.to_yaml()
        model_data_attrs['_debug_data'] = model._debug_data.to_yaml()

    # Convert boolean attrs to ints
    bool_attrs = [
        k for k, v in model_data_attrs.items()
        if isinstance(v, bool)
    ]
    for k in bool_attrs:
        model_data_attrs[k] = int(model_data_attrs[k])

    # Convert None attrs to 'None'
    none_attrs = [
        k for k, v in model_data_attrs.items()
        if v is None
    ]
    for k in none_attrs:
        model_data_attrs[k] = 'None'

    # Convert `object` dtype coords to string
    # FIXME: remove once xarray issue https://github.com/pydata/xarray/issues/2404 is resolved
    for k, v in model_data.coords.items():
        if v.dtype == 'O':
            model_data[k] = v.astype('<U{}'.format(max([len(i.item()) for i in v])))

    try:
        model_data.attrs = model_data_attrs
        model_data.to_netcdf(path, format='netCDF4', encoding=encoding)
        model_data.close()  # Force-close NetCDF file after writing
    finally:  # Revert model_data.attrs back
        model_data.attrs = original_model_data_attrs


def save_csv(model_data, path, dropna=True):
    """
    If termination condition was not optimal, filters inputs only, and
    warns that results will not be saved.

    """
    os.makedirs(path, exist_ok=False)

    # a MILP model which optimises to within the MIP gap, but does not fully
    # converge on the LP relaxation, may return as 'feasible', not 'optimal'
    if ('termination_condition' not in model_data.attrs or
            model_data.attrs['termination_condition'] in ['optimal', 'feasible']):
        data_vars = model_data.data_vars
    else:
        data_vars = model_data.filter_by_attrs(is_result=0).data_vars
        exceptions.warn(
            'Model termination condition was not optimal, saving inputs only.'
        )

    for var in data_vars:
        in_out = 'results' if model_data[var].attrs['is_result'] else 'inputs'
        out_path = os.path.join(path, '{}_{}.csv'.format(in_out, var))
        series = split_loc_techs(model_data[var], return_as='Series')
        if dropna:
            series = series.dropna()
        series.to_csv(out_path, header=True)


def save_lp(model, path):
    if not model.run_config['backend'] == 'pyomo':
        raise IOError('Only the pyomo backend can save to LP.')
    if not hasattr(model, '_backend_model'):
        model.run(build_only=True)
    model._backend_model.write(path, format='lp', io_options={'symbolic_solver_labels': True})



import shutil

import pytest
from pytest import approx
import pandas as pd

import calliope
from calliope.test.common.util import check_error_or_warning


class TestModelPreproccesing:
    def test_preprocess_national_scale(self):
        calliope.examples.national_scale()

    def test_preprocess_time_clustering(self):
        calliope.examples.time_clustering()

    def test_preprocess_time_resampling(self):
        calliope.examples.time_resampling()

    def test_preprocess_urban_scale(self):
        calliope.examples.urban_scale()

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_preprocess_milp(self):
        calliope.examples.milp()

    def test_preprocess_operate(self):
        calliope.examples.operate()

    def test_preprocess_time_masking(self):
        calliope.examples.time_masking()


class TestNationalScaleExampleModelSenseChecks:
    def example_tester(self, solver='cbc', solver_io=None):
        override = {
            'model.subset_time': '2005-01-01',
            'run.solver': solver,
        }

        if solver_io:
            override['run.solver_io'] = solver_io

        model = calliope.examples.national_scale(override_dict=override)
        model.run()

        assert model.results.storage_cap.to_pandas()['region1-1::csp'] == approx(45129.950)
        assert model.results.storage_cap.to_pandas()['region2::battery'] == approx(6675.173)

        assert model.results.energy_cap.to_pandas()['region1-1::csp'] == approx(4626.588)
        assert model.results.energy_cap.to_pandas()['region2::battery'] == approx(1000)
        assert model.results.energy_cap.to_pandas()['region1::ccgt'] == approx(30000)

        assert float(model.results.cost.sum()) == approx(38988.7442)

        assert float(
            model.results.systemwide_levelised_cost.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.063543, abs=0.000001)
        assert float(
            model.results.systemwide_capacity_factor.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.2642256, abs=0.000001)

    def test_nationalscale_example_results_cbc(self):
        self.example_tester()

    def test_nationalscale_example_results_gurobi(self):
        try:
            import gurobipy  # pylint: disable=unused-import
            self.example_tester(solver='gurobi', solver_io='python')
        except ImportError:
            pytest.skip('Gurobi not installed')

    def test_nationalscale_example_results_cplex(self):
        if shutil.which('cplex'):
            self.example_tester(solver='cplex')
        else:
            pytest.skip('CPLEX not installed')

    def test_nationalscale_example_results_glpk(self):
        if shutil.which('glpsol'):
            self.example_tester(solver='glpk')
        else:
            pytest.skip('GLPK not installed')

    def test_considers_supply_generation_only_in_total_levelised_cost(self):
        # calculation of expected value:
        # costs = model.get_formatted_array("cost").sum(dim="locs")
        # gen = model.get_formatted_array("carrier_prod").sum(dim=["timesteps", "locs"])
        # lcoe = costs.sum(dim="techs") / gen.sel(techs=["ccgt", "csp"]).sum(dim="techs")
        model = calliope.examples.national_scale()
        model.run()

        assert model.results.total_levelised_cost.item() == approx(0.067005, abs=1e-5)

    def test_fails_gracefully_without_timeseries(self):
        override = {
            "locations.region1.techs.demand_power.constraints.resource": -200,
            "locations.region2.techs.demand_power.constraints.resource": -400,
            "techs.csp.constraints.resource": 100
        }
        with pytest.raises(calliope.exceptions.ModelError):
            calliope.examples.national_scale(override_dict=override)


class TestNationalScaleExampleModelInfeasibility:
    def example_tester(self):

        model = calliope.examples.national_scale(
            scenario='check_feasibility',
            override_dict={'run.cyclic_storage': False}
        )

        model.run()

        assert model.results.attrs['termination_condition'] in ['infeasible', 'other']  # glpk gives 'other' as result

        assert 'systemwide_levelised_cost' not in model.results.data_vars
        assert 'systemwide_capacity_factor' not in model.results.data_vars

    def test_nationalscale_example_results_cbc(self):
        self.example_tester()


class TestNationalScaleExampleModelOperate:
    def example_tester(self):
        with pytest.warns(calliope.exceptions.ModelWarning) as excinfo:
            model = calliope.examples.national_scale(
                override_dict={'model.subset_time': ['2005-01-01', '2005-01-03']},
                scenario='operate')
            model.run()

        expected_warning = 'Resource capacity constraint defined and set to infinity for all supply_plus techs'

        assert check_error_or_warning(excinfo, expected_warning)
        assert all(model.results.timesteps == pd.date_range('2005-01', '2005-01-03 23:00:00', freq='H'))

    def test_nationalscale_example_results_cbc(self):
        self.example_tester()


class TestNationalScaleResampledExampleModelSenseChecks:
    def example_tester(self, solver='cbc', solver_io=None):
        override = {
            'model.subset_time': '2005-01-01',
            'run.solver': solver,
        }

        if solver_io:
            override['run.solver_io'] = solver_io

        model = calliope.examples.time_resampling(override_dict=override)
        model.run()

        assert model.results.storage_cap.to_pandas()['region1-1::csp'] == approx(23563.444)
        assert model.results.storage_cap.to_pandas()['region2::battery'] == approx(6315.78947)

        assert model.results.energy_cap.to_pandas()['region1-1::csp'] == approx(1440.8377)
        assert model.results.energy_cap.to_pandas()['region2::battery'] == approx(1000)
        assert model.results.energy_cap.to_pandas()['region1::ccgt'] == approx(30000)

        assert float(model.results.cost.sum()) == approx(37344.221869)

        assert float(
            model.results.systemwide_levelised_cost.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.063543, abs=0.000001)
        assert float(
            model.results.systemwide_capacity_factor.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.25, abs=0.000001)

    def test_nationalscale_resampled_example_results_cbc(self):
        self.example_tester()

    def test_nationalscale_resampled_example_results_glpk(self):
        if shutil.which('glpsol'):
            self.example_tester(solver='glpk')
        else:
            pytest.skip('GLPK not installed')


class TestNationalScaleClusteredExampleModelSenseChecks:
    def model_runner(self, solver='cbc', solver_io=None,
                     how='closest', storage_inter_cluster=False,
                     cyclic=False, storage=True):
        override = {
            'model.time.function_options': {
                'how': how, 'storage_inter_cluster': storage_inter_cluster
            },
            'run.solver': solver,
            'run.cyclic_storage': cyclic
        }
        if storage is False:
            override.update({
                'techs.battery.exists': False,
                'techs.csp.exists': False
            })

        if solver_io:
            override['run.solver_io'] = solver_io

        model = calliope.examples.time_clustering(override_dict=override)
        model.run()

        return model

    def example_tester_closest(self, solver='cbc', solver_io=None):
        model = self.model_runner(solver=solver, solver_io=solver_io, how='closest')
        # Full 1-hourly model run: 22312488.670967
        assert float(model.results.cost.sum()) == approx(49670627.15297682)

        # Full 1-hourly model run: 0.296973
        assert float(
            model.results.systemwide_levelised_cost.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.137105, abs=0.000001)

        # Full 1-hourly model run: 0.064362
        assert float(
            model.results.systemwide_capacity_factor.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.064501, abs=0.000001)

    def example_tester_mean(self, solver='cbc', solver_io=None):
        model = self.model_runner(solver=solver, solver_io=solver_io, how='mean')
        # Full 1-hourly model run: 22312488.670967
        assert float(model.results.cost.sum()) == approx(22172253.328)

        # Full 1-hourly model run: 0.296973
        assert float(
            model.results.systemwide_levelised_cost.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.127783, abs=0.000001)

        # Full 1-hourly model run: 0.064362
        assert float(
            model.results.systemwide_capacity_factor.loc[dict(carriers='power')].to_pandas().T['battery']
        ) == approx(0.044458, abs=0.000001)

    def example_tester_storage_inter_cluster(self):
        model = self.model_runner(storage_inter_cluster=True)

        # Full 1-hourly model run: 22312488.670967
        assert float(model.results.cost.sum()) == approx(21825515.304)

        # Full 1-hourly model run: 0.296973
        assert float(
            model.results.systemwide_levelised_cost.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.100760, abs=0.000001)

        # Full 1-hourly model run: 0.064362
        assert float(
            model.results.systemwide_capacity_factor.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.091036, abs=0.000001)

    def test_nationalscale_clustered_example_closest_results_cbc(self):
        self.example_tester_closest()

    def test_nationalscale_clustered_example_closest_results_glpk(self):
        if shutil.which('glpsol'):
            self.example_tester_closest(solver='glpk')
        else:
            pytest.skip('GLPK not installed')

    def test_nationalscale_clustered_example_mean_results_cbc(self):
        self.example_tester_mean()

    @pytest.mark.skip(reason='GLPK is useless and delivering different results on different operating systems')
    def test_nationalscale_clustered_example_mean_results_glpk(self):
        if shutil.which('glpsol'):
            self.example_tester_mean(solver='glpk')
        else:
            pytest.skip('GLPK not installed')

    def test_nationalscale_clustered_example_storage_inter_cluster(self):
        self.example_tester_storage_inter_cluster()

    def test_storage_inter_cluster_cyclic(self):
        model = self.model_runner(storage_inter_cluster=True, cyclic=True)
        # Full 1-hourly model run: 22312488.670967
        assert float(model.results.cost.sum()) == approx(18904055.722)

        # Full 1-hourly model run: 0.296973
        assert float(
            model.results.systemwide_levelised_cost.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.122564, abs=0.000001)

        # Full 1-hourly model run: 0.064362
        assert float(
            model.results.systemwide_capacity_factor.loc[{'carriers': 'power', 'techs': 'battery'}].item()
        ) == approx(0.075145, abs=0.000001)

    def test_storage_inter_cluster_no_storage(self):
        with pytest.warns(calliope.exceptions.ModelWarning) as excinfo:
            self.model_runner(storage_inter_cluster=True, storage=False)

        expected_warnings = [
            'Tech battery was removed by setting ``exists: False``',
            'Tech csp was removed by setting ``exists: False``'
        ]
        assert check_error_or_warning(excinfo, expected_warnings)


class TestUrbanScaleExampleModelSenseChecks:
    def example_tester(self, resource_unit, solver='cbc', solver_io=None):
        unit_override = {
            'techs.pv.constraints': {
                'resource': 'file=pv_resource.csv:{}'.format(resource_unit),
                'resource_unit': 'energy_{}'.format(resource_unit)
            },
            'run.solver': solver
        }
        override = {'model.subset_time': '2005-07-01', **unit_override}

        if solver_io:
            override['run.solver_io'] = solver_io

        model = calliope.examples.urban_scale(override_dict=override)
        model.run()

        assert model.results.energy_cap.to_pandas()['X1::chp'] == approx(250.090112)

        # GLPK isn't able to get the same answer both times, so we have to account for that here
        if resource_unit == 'per_cap' and solver == 'glpk':
            heat_pipe_approx = 183.45825
        else:
            heat_pipe_approx = 182.19260

        assert model.results.energy_cap.to_pandas()['X2::heat_pipes:N1'] == approx(heat_pipe_approx)

        assert model.results.carrier_prod.sum('timesteps').to_pandas()['X3::boiler::heat'] == approx(0.18720)
        assert model.results.resource_area.to_pandas()['X2::pv'] == approx(830.064659)

        assert float(model.results.carrier_export.sum()) == approx(122.7156)

        # GLPK doesn't agree with commercial solvers, so we have to account for that here
        cost_sum = 430.097399 if solver == 'glpk' else 430.089188
        assert float(model.results.cost.sum()) == approx(cost_sum)

    def test_urban_example_results_area(self):
        self.example_tester('per_area')

    def test_urban_example_results_area_gurobi(self):
        try:
            import gurobipy  # pylint: disable=unused-import
            self.example_tester('per_area', solver='gurobi', solver_io='python')
        except ImportError:
            pytest.skip('Gurobi not installed')

    def test_urban_example_results_cap(self):
        self.example_tester('per_cap')

    def test_urban_example_results_cap_gurobi(self):
        try:
            import gurobipy  # pylint: disable=unused-import
            self.example_tester('per_cap', solver='gurobi', solver_io='python')
        except ImportError:
            pytest.skip('Gurobi not installed')

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_milp_example_results(self):
        model = calliope.examples.milp(
            override_dict={'model.subset_time': '2005-01-01', 'run.solver_options.mipgap': 0.001}
        )
        model.run()

        assert model.results.energy_cap.to_pandas()['X1::chp'] == 300
        assert model.results.energy_cap.to_pandas()['X2::heat_pipes:N1'] == approx(188.363137)

        assert model.results.carrier_prod.sum('timesteps').to_pandas()['X1::supply_gas::gas'] == approx(12363.173036)
        assert float(model.results.carrier_export.sum()) == approx(0)

        assert model.results.purchased.to_pandas()['X2::boiler'] == 1
        assert model.results.units.to_pandas()['X1::chp'] == 1

        assert float(model.results.operating_units.sum()) == 24

        assert float(model.results.cost.sum()) == approx(540.780779)

    def test_operate_example_results(self):
        model = calliope.examples.operate(
            override_dict={'model.subset_time': ['2005-07-01', '2005-07-04']}
        )
        with pytest.warns(calliope.exceptions.ModelWarning) as excinfo:
            model.run()

        expected_warnings = [
            'Energy capacity constraint removed',
            'Resource capacity constraint defined and set to infinity for all supply_plus techs',
            'Storage cannot be cyclic in operate run mode, setting `run.cyclic_storage` to False for this run'
        ]

        assert check_error_or_warning(excinfo, expected_warnings)

        assert all(model.results.timesteps == pd.date_range('2005-07', '2005-07-04 23:00:00', freq='H'))



from calliope._version import __version__

from calliope import examples, exceptions
from calliope.core import AttrDict, Model, read_netcdf
from calliope.core.util.logging import set_log_verbosity


__title__ = 'Calliope'
__author__ = 'Calliope contributors listed in AUTHORS'
__copyright__ = 'Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS'



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

plotting.py
~~~~~~~~~~~

Functionality to plot model data.

"""

import os
import re

import plotly.offline as pltly
import jinja2

from calliope.exceptions import warn
from calliope.analysis.plotting.capacity import plot_capacity
from calliope.analysis.plotting.timeseries import plot_timeseries
from calliope.analysis.plotting.transmission import plot_transmission
from calliope.analysis.plotting.flows import plot_flows
from calliope.analysis.plotting.util import type_of_script


def plot_summary(model, to_file=False, mapbox_access_token=None):
    """
    Plot a summary containing timeseries, installed capacities, and
    transmission plots. Returns a HTML string by default, returns None if
    ``to_file`` given (and saves the HTML string to file).

    Parameters
    ----------
    to_file : str, optional
        Path to output file to save HTML to.
    mapbox_access_token : str, optional
        (passed to plot_transmission) If given and a valid Mapbox API
        key, a Mapbox map is drawn for lat-lon coordinates, else
        (by default), a more simple built-in map.

    """
    subset = {'costs': ['monetary']}

    timeseries = _plot(*plot_timeseries(model, subset=subset), html_only=True)
    capacity = _plot(*plot_capacity(model, subset=subset), html_only=True)
    if 'loc_coordinates' in model._model_data:
        transmission = _plot(*plot_transmission(
            model, html_only=True, mapbox_access_token=mapbox_access_token
        ), html_only=True)
    else:
        transmission = '<br><br><p>No location coordinates defined -<br>not plotting transmission.</p>'

    template_path = os.path.join(
        os.path.dirname(__file__), '..', '..', 'config', 'plots_template.html'
    )
    with open(template_path, 'r') as f:
        html_template = jinja2.Template(f.read())

    if 'solution_time' in model._model_data.attrs:
        solution_time = model._model_data.attrs['solution_time'] / 60
        time_finished = model._model_data.attrs['time_finished']
        result_stats = 'taking {:.2f} minutes to solve, completed at {}'.format(
            solution_time, time_finished
        )
    else:
        result_stats = 'inputs only'

    html = html_template.render(
        model_name=model.model_config['name'],
        calliope_version=model._model_data.attrs['calliope_version'],
        result_stats=result_stats,
        top=timeseries,
        bottom_left=capacity,
        bottom_right=transmission,
    )

    # Strip plotly-inserted style="..." attributes
    html = re.sub(r'style=".+?"', '', html)

    if to_file:
        with open(to_file, 'w') as f:
            f.write(html)
    else:
        return html


def _plot(
        data, layout, html_only=False, to_file=False,
        layout_updates=None, plotly_kwarg_updates=None,
        # kwargs are included as they get passed through from the
        # plotting accessor method, but are not actually used
        **kwargs):

    plotly_kwargs = dict(
        show_link=False,
        config={
            'displaylogo': False,
            'modeBarButtonsToRemove': ['sendDataToCloud'],
        },
    )

    if type_of_script() == 'jupyter':
        plotter = pltly.iplot
        plotly_filename_key = 'filename'
        pltly.init_notebook_mode(connected=False)
    else:
        plotter = pltly.plot
        plotly_filename_key = 'image_filename'
        plotly_kwargs['auto_open'] = True
        plotly_kwargs['filename'] = 'temp_plot.html'

    if layout_updates:
        layout = {**layout, **layout_updates}

    if plotly_kwarg_updates:
        plotly_kwargs = {**plotly_kwargs, **plotly_kwarg_updates}

    if to_file:
        filename, image_type = to_file.rsplit('.', 1)
        # Plotly can only save certain file types
        if image_type not in ['png', 'jpeg', 'svg', 'webp']:
            raise TypeError(
                'Unable to save plot as `{}`, choose from '
                '[`png`, `jpeg`, `svg`, `webp`]'.format(image_type)
            )
        if 'updatemenus' in layout:
            raise ValueError('Unable to save multiple arrays to SVG, pick one array only')
        else:
            plotly_kwargs.update(image=image_type)
            plotly_kwargs[plotly_filename_key] = filename

    if data:
        if html_only:
            return pltly.plot(
                {'data': data, 'layout': layout},
                include_plotlyjs=False, output_type='div',
                **plotly_kwargs
            )
        else:
            plotter({'data': data, 'layout': layout}, **plotly_kwargs)
    else:
        raise ValueError('No data to plot.')


class ModelPlotMethods:
    def __init__(self, model):
        self._model = model

    _docstring_additions = """
    html_only : bool, optional; default = False
        Returns a html string for embedding the plot in a webpage
    to_file : False or str, optional; default = False
        Will save plot to file with the given name and extension.
        `to_file='plot.svg'` to save to SVG, `to_file='plot.png'` for
        a static PNG image.
        Allowed file extensions are: ['png', 'jpeg', 'svg', 'webp'].
    layout_updates : dict, optional
        The given dict will be merged with the Plotly layout dict
        generated by the Calliope plotting function, overwriting keys
        that already exist.
    plotly_kwarg_updates : dict, optional
        The given dict will be merged with the Plotly plot function's
        keyword arguments generated by the Calliope plotting function,
        overwriting keys that already exist.

    """

    def check_optimality(self):
        termination = self._model._model_data.attrs.get(
            'termination_condition', 'did_not_yet_run')
        # a MILP model which optimises to within the MIP gap, but does not fully
        # converge on the LP relaxation, may return as 'feasible', not 'optimal'
        if termination not in ['optimal', 'did_not_yet_run', 'feasible']:
            warn('Model termination condition was not optimal. Plotting may fail!')

    def timeseries(self, **kwargs):
        self.check_optimality()
        data, layout = plot_timeseries(self._model, **kwargs)
        return _plot(data, layout, **kwargs)

    timeseries.__doc__ = plot_timeseries.__doc__.rstrip() + _docstring_additions

    def capacity(self, **kwargs):
        self.check_optimality()
        data, layout = plot_capacity(self._model, **kwargs)
        return _plot(data, layout, **kwargs)

    capacity.__doc__ = plot_capacity.__doc__.rstrip() + _docstring_additions

    def transmission(self, **kwargs):
        self.check_optimality()
        data, layout = plot_transmission(self._model, **kwargs)
        return _plot(data, layout, **kwargs)

    transmission.__doc__ = plot_transmission.__doc__.rstrip() + _docstring_additions

    def flows(self, **kwargs):
        self.check_optimality()
        data, layout = plot_flows(self._model, **kwargs)
        return _plot(data, layout, **kwargs)

    flows.__doc__ = plot_flows.__doc__.rstrip() + _docstring_additions

    def summary(self, **kwargs):
        self.check_optimality()
        return plot_summary(self._model, **kwargs)

    summary.__doc__ = plot_summary.__doc__
